/*
Copyright (c) 2010, Stephane Duguay < s (@) binarez (.) com >
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this list of conditions and
the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions
and the following disclaimer in the documentation and/or other materials provided with the distribution.
* Neither the name of the <ORGANIZATION> nor the names of its contributors may be used to endorse or promote
products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

//------------------------------------------------------------------------------------------------//
#include "filecrawler.h"
#include "touchutils.h"
//------------------------------------------------------------------------------------------------//
static const std::wstring kBackslash = L"\\";
static const std::wstring kBackslashStar = L"\\*";
//------------------------------------------------------------------------------------------------//
FileCrawler::FileCrawler( const std::wstring& absTarget, bool recursive )
:	mTargetAbs( absTarget )
,	mhFindFiles( INVALID_HANDLE_VALUE )
,	mRecursive( recursive )
,	mpActiveSubCrawler( NULL )
{
	memset( &mCurrentFileInfo, 0, sizeof( mCurrentFileInfo ) );
	if( !mTargetAbs.empty() )
	{
		TouchUtils::removeTrailingSlashes( mTargetAbs );
		// If you are recursively targeting a directory, we'll automatically add '\*' to help the user.
		if( recursive && ( TouchUtils::getPathType( mTargetAbs ) == TouchUtils::ePathType::Directory ) )
		{
			mTargetAbs.append( kBackslashStar );
		}
		const std::wstring::size_type foundPos = mTargetAbs.rfind( L'\\' );
		if( foundPos != std::wstring::npos )
		{
			mCurrentAbsDir = mTargetAbs.substr( 0, foundPos );
			mCurrentFilter = mTargetAbs.substr( foundPos + 1 );

			if( recursive )
			{
				detectSubDirectories( mCurrentAbsDir, mCurrentFilter );
			}

			mhFindFiles = FindFirstFile( mTargetAbs.c_str(), &mCurrentFileInfo );
			if( mhFindFiles != INVALID_HANDLE_VALUE )
			{
				// Skip useless directories
				while(	mCurrentFileInfo.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY	&&
						( wcscmp( mCurrentFileInfo.cFileName, L".." ) == 0 || wcscmp( mCurrentFileInfo.cFileName, L"." ) == 0 ) )
				{
					if( FindNextFile( mhFindFiles, &mCurrentFileInfo ) == 0 )
					{
						die();
						return;
					}
				}
			}
			else
			{
				nextFile();
			}

			// Success
			return;
		}
	}
	die();
}
//------------------------------------------------------------------------------------------------//
FileCrawler::~FileCrawler( void )
{
	die();

	// Delete subcrawlers
	delete mpActiveSubCrawler;
	while( !mSubCrawlers.empty() )
	{
		delete mSubCrawlers.front();
		mSubCrawlers.pop();
	}
}
//------------------------------------------------------------------------------------------------//
void FileCrawler::getCurrentFilename( std::wstring& absFilename ) const
{
	absFilename.resize( 0 );	// This preserves memory on most STL implementations
	if( mhFindFiles != INVALID_HANDLE_VALUE )
	{
		absFilename.reserve( mCurrentAbsDir.length() + kBackslash.length() + wcslen( mCurrentFileInfo.cFileName ) );
		absFilename.append( mCurrentAbsDir );
		absFilename.append( kBackslash );
		absFilename.append( mCurrentFileInfo.cFileName );
	}
	else if( mpActiveSubCrawler != NULL )
	{
		mpActiveSubCrawler->getCurrentFilename( absFilename );
	}
}
//------------------------------------------------------------------------------------------------//
bool FileCrawler::nextFile( void )
{
	do
	{
		if( !nextFileInternal() )
		{
			if( !nextFileSubCrawl() )
			{
				return false;
			}
		}
	} while( mCurrentFileInfo.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY );
	return gotFiles();
}
//------------------------------------------------------------------------------------------------//
bool FileCrawler::nextFileInternal( void )
{
	if( mhFindFiles != NULL )
	{
		if( FindNextFile( mhFindFiles, &mCurrentFileInfo ) != 0 )
			{ return true; }
		else
			{ die(); }
	}
	return false;
}
//------------------------------------------------------------------------------------------------//
bool FileCrawler::nextFileSubCrawl( void )
{
	if( mpActiveSubCrawler != NULL )
	{
		const bool subOK = mpActiveSubCrawler->nextFile();
		if( subOK )
		{
			return true;
		}
		else
		{
			delete mpActiveSubCrawler;
			mpActiveSubCrawler = NULL;
		}
	}

	if( mpActiveSubCrawler == NULL && !mSubCrawlers.empty() )
	{
		mpActiveSubCrawler = mSubCrawlers.front();
		mSubCrawlers.pop();
	}

	return mpActiveSubCrawler != NULL;
}
//------------------------------------------------------------------------------------------------//
void FileCrawler::die( void )
{
	if( mhFindFiles != INVALID_HANDLE_VALUE )
	{
		FindClose( mhFindFiles );
		mhFindFiles = INVALID_HANDLE_VALUE;
	}
	//mCurrentAbsDir.resize( 0 );
	//mCurrentFilter.resize( 0 );
	memset( &mCurrentFileInfo, 0, sizeof( mCurrentFileInfo ) );
}
//------------------------------------------------------------------------------------------------//
bool FileCrawler::gotFiles( void ) const
{
	return	mhFindFiles != INVALID_HANDLE_VALUE	||
			mpActiveSubCrawler != NULL			||
			!mSubCrawlers.empty();
}
//------------------------------------------------------------------------------------------------//
void FileCrawler::detectSubDirectories( const std::wstring& currentAbsDir, const std::wstring& currentFilter )
{
	WIN32_FIND_DATA dirsFindData;
	std::wstring searchFilter;
	searchFilter.reserve( currentAbsDir.length() + kBackslashStar.length() );
	searchFilter.append( currentAbsDir );
	searchFilter.append( kBackslashStar );
	HANDLE hSubDirs = FindFirstFile( searchFilter.c_str(), &dirsFindData );
	if( hSubDirs != INVALID_HANDLE_VALUE )
	{
		do 
		{
			// If we have a directory other than '.' or '..'
			if(	( dirsFindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY )	&&
				( wcscmp( dirsFindData.cFileName, L"." ) != 0 && wcscmp( dirsFindData.cFileName, L".." ) != 0 ) )
			{
				std::wstring absDirPath;
				absDirPath.reserve( currentAbsDir.length() + kBackslash.length() + wcslen( dirsFindData.cFileName ) );
				absDirPath.append( currentAbsDir );
				absDirPath.append( kBackslash );
				absDirPath.append( dirsFindData.cFileName );
				FileCrawler* pSubCrawler = new FileCrawler( absDirPath + kBackslash + currentFilter, mRecursive );
				if( pSubCrawler != NULL )
				{
					// Don't bother with empty directories
					if( pSubCrawler->gotFiles() )
					{
						mSubCrawlers.push( pSubCrawler );
					}
					else
					{
						delete pSubCrawler;
						pSubCrawler = NULL;
					}
				}
			}
		} while ( FindNextFile( hSubDirs, &dirsFindData ) != 0 );

		FindClose( hSubDirs );
	}
}
//------------------------------------------------------------------------------------------------//
