/*
Copyright (c) 2010, Stephane Duguay < s (@) binarez (.) com >
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this list of conditions and
the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions
and the following disclaimer in the documentation and/or other materials provided with the distribution.
* Neither the name of the <ORGANIZATION> nor the names of its contributors may be used to endorse or promote
products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

//------------------------------------------------------------------------------------------------//
#include "touchoptions.h"
#include "touchtime.h"
#include "filecrawler.h"
#include "touchutils.h"
//------------------------------------------------------------------------------------------------//
#include <iostream>
#include <windows.h>
//------------------------------------------------------------------------------------------------//
#define TOUCH_EXE_MAJOR_VERSION		0
#define TOUCH_EXE_MINOR_VERSION		99
//------------------------------------------------------------------------------------------------//
int processFiles( int argc, wchar_t* argv[], const TouchOptions& touchOptions, const TouchTime& touchTime );
bool processSingleFile(	const std::wstring& absFilename, const TouchOptions& touchOptions,
						const TouchTime& touchTime, std::vector< std::wstring >& errors );
bool touchFile( const std::wstring& filename, const TouchOptions& touchOptions, const TouchTime& touchTime );
void getAbsolutePath( const std::wstring& relativePath, std::wstring& absPath );
void printUsage( void );
void printOptionUsage( const wchar_t* option, const wchar_t* desc );
void printVersion( void );
void printStrings( const std::vector< std::wstring >& toPrint );
void printClosingMessage( bool gotSuccess, const std::vector< std::wstring >& errors, int exitCode );
void printUnicodeSafely( const std::wstring& filename );
void printTimeLoadErrorMsg( TouchTime::eLoadResult timeLoadResult );
//------------------------------------------------------------------------------------------------//
static const std::wstring kProgramName = L"touch";
static const std::wstring kLink = L":\t";
static const std::wstring kUnknownError = L"Error";
static const std::wstring kCannotTouchDirectoryError =	L"Cannot touch a directory.\n"
														L"*Tip* 'cd' into that directory then 'touch *' or use --recursive mode\n"
														L"       to touch files in specified directory and all subdirectories.";
static const std::wstring kCreatedMsg = L" (Created)";
static const std::wstring kReadOnlyError = L"The file is read-only.";
static const std::wstring kQuestionMark = L"?";
//------------------------------------------------------------------------------------------------//
int wmain( int argc, wchar_t* argv[] )
{
	int exitCode = 0;

	// With this code, std::wcout will throw an exception when it can't print a unicode character
	// instead of just going into a bad-state and not outputing anymore. See the printUnicodeSafely function.
	std::wcout.exceptions( std::ios::badbit );

	// Setup locale, console programs default to "C" locale.
	const char * const currentLocale = setlocale( LC_CTYPE, NULL );
	setlocale( LC_CTYPE, "" );

	// Load touch options
	int filesIndex = 0;
	TouchOptions touchOptions( argc, argv, filesIndex );

	if( touchOptions.mHelp )
	{
		printUsage();
	}
	else if( touchOptions.mVersion )
	{
		printVersion();
	}
	else if( filesIndex > 0 && filesIndex < argc )
	{
		if( !touchOptions.mReferenceFile.empty() )
		{
			std::wstring absRefFile;
			getAbsolutePath( touchOptions.mReferenceFile, absRefFile );
			touchOptions.mReferenceFile = absRefFile;
		}

		TouchTime touchTime;
		const TouchTime::eLoadResult timeLoadResult = touchTime.loadFromOptions( touchOptions );
		if( timeLoadResult == TouchTime::Success )
		{
			exitCode = processFiles( argc - filesIndex, argv + filesIndex, touchOptions, touchTime );
		}
		else
		{
			printTimeLoadErrorMsg( timeLoadResult );
			exitCode = -2;
		}
	}
	else
	{
		std::wcout << kProgramName << L": missing FILE operand." << std::endl;
		std::wcout << L"Type '" << kProgramName << " --help' for more information." << std::endl;
		exitCode = -1;
	}

	if( touchOptions.mPauseOnExit )
		{ ::system( "pause" ); }
	
	setlocale( LC_CTYPE, currentLocale );

	return exitCode;
}
//------------------------------------------------------------------------------------------------//
// This function returns the number of files that triggered an error.
int processFiles( int argc, wchar_t* argv[], const TouchOptions& touchOptions, const TouchTime& touchTime )
{
	if( touchOptions.mVerbose )
	{
		std::wcout << std::endl << L"Successful:" << std::endl;
	}
	std::vector< std::wstring > errors;
	int numErrors = 0;
	bool gotSuccess = false;
	for( int f = 0; f < argc; ++f )
	{
		static std::wstring targetAbsPath; // static prevents realloc
		getAbsolutePath( argv[f], targetAbsPath );
		FileCrawler finder( targetAbsPath, touchOptions.mRecursive );
		if( finder.gotFiles() )
		{
			// Found files, let's crawl.
			do
			{
				static std::wstring currentFileAbs;	// static prevents realloc
				finder.getCurrentFilename( currentFileAbs );
				if( processSingleFile( currentFileAbs, touchOptions, touchTime, errors ) )
				{
					gotSuccess = true;
					if( touchOptions.mVerbose )
					{
						printUnicodeSafely( currentFileAbs );
						std::wcout << std::endl;
					}
				}
				else
				{
					numErrors++;
				}
			} while( finder.nextFile() );
		}
		// This file doesn't exist, should it be created?
		else if( ( TouchUtils::getPathType( targetAbsPath ) != TouchUtils::ePathType::Search )	&&
				 ( TouchUtils::getPathType( targetAbsPath ) != TouchUtils::ePathType::Directory ) )
		{
			if( touchOptions.mNoCreate )
			{
				if( touchOptions.mVerbose )
					{ errors.push_back( std::wstring( argv[f] ) + kLink + L"File not created because of --no-create." ); }
				numErrors++;
			}
			else
			{
				TouchOptions creationTouchOptions( touchOptions );
				creationTouchOptions.mChangeCreationTime = true;
				creationTouchOptions.mChangeAccessTime = true;
				creationTouchOptions.mChangeModifTime = true;
				if( touchFile( targetAbsPath, creationTouchOptions, touchTime ) )
				{
					gotSuccess = true;
					if( touchOptions.mVerbose )
					{
						printUnicodeSafely( targetAbsPath );
						std::wcout << kCreatedMsg;
						std::wcout << std::endl;
					}
				}
				else
				{
					if( touchOptions.mVerbose )
						{ errors.push_back( std::wstring( argv[f] ) + kLink + L"Cannot create file." ); }
					numErrors++;
				}
			}
		}
		else	// Empty dir
		{
		}
	}

	if( touchOptions.mVerbose )
	{
		printClosingMessage( gotSuccess, errors, numErrors );
	}

	return numErrors;
}
//------------------------------------------------------------------------------------------------//
bool processSingleFile( const std::wstring& absFilename, const TouchOptions& touchOptions,
						const TouchTime& touchTime, std::vector< std::wstring >& errors )
{
	if( absFilename.empty()	||
		!touchFile( absFilename, touchOptions, touchTime ) )
	{
		if( touchOptions.mVerbose )
		{
			std::wstring thisError;
			if( absFilename.empty() )
				{ thisError.append( kUnknownError ); }
			else if( TouchUtils::getPathType( absFilename ) == TouchUtils::ePathType::Directory && !touchOptions.mRecursive )
			{
				thisError.append( absFilename );
				thisError.append( kLink );
				thisError.append( kCannotTouchDirectoryError );
			}
			else
			{
				thisError.append( absFilename );
				if( GetFileAttributes( absFilename.c_str() ) & FILE_ATTRIBUTE_READONLY )
				{
					thisError.append( kLink );
					thisError.append( kReadOnlyError );
				}
			}
			errors.push_back( thisError );
		}
		return false;
	}

	return true;
}
//------------------------------------------------------------------------------------------------//
bool touchFile( const std::wstring& filename, const TouchOptions& touchOptions, const TouchTime& touchTime )
{
	HANDLE hFile = CreateFile(	filename.c_str(), GENERIC_WRITE, 0, NULL,
								touchOptions.mNoCreate ? OPEN_EXISTING : OPEN_ALWAYS,
								FILE_ATTRIBUTE_NORMAL, NULL );
	if( hFile != INVALID_HANDLE_VALUE )	// CreateFile success!
	{
		SetFileTime(	hFile,
						touchOptions.mChangeCreationTime	? &touchTime.mCreationTime	: NULL,
						touchOptions.mChangeAccessTime		? &touchTime.mAccessTime	: NULL,
						touchOptions.mChangeModifTime		? &touchTime.mModifTime		: NULL	);
		CloseHandle( hFile );
		return true;
	}
	return false;
}
//------------------------------------------------------------------------------------------------//
void getAbsolutePath( const std::wstring& relativePath, std::wstring& absPath )
{
	if( relativePath.empty() )
	{
		absPath.resize(0);
		return;
	}

	const int kBufferSize = 1024;
	absPath.resize( kBufferSize, ' ' );
	DWORD lenFullPath = 0;
	do
	{
		absPath.resize( absPath.size() * 2 );
		lenFullPath = GetFullPathName( relativePath.c_str(), static_cast< DWORD >( absPath.size() ), &absPath[0], NULL );
	} while( lenFullPath > absPath.size() );
	absPath.resize( lenFullPath );
}
//------------------------------------------------------------------------------------------------//
void printUsage()
{
	std::wcout << L"Changes files access, modification and creation times." << std::endl;
	std::wcout << std::endl;
	std::wcout << L"Usage: " << kProgramName << L" [-acmpsRvx] [-r REFFILE | -t TIME | -d DATETIME] FILE..." << std::endl;
	std::wcout << std::endl;
	std::wcout << L"A FILE argument that does not exist is created empty, unless -c or --no-create" << std::endl;
	std::wcout << L"is supplied." << std::endl;
	std::wcout << std::endl;

	printOptionUsage( L"-a, --access-time      ", L"Change file access time." );
	printOptionUsage( L"-c, --no-create        ", L"Do not create any new files." );
	printOptionUsage( L"-m, --modif-time       ", L"Change file modification time." );
	printOptionUsage( L"-p, --pause-exit       ", L"Pause on exit." );
	printOptionUsage( L"-s, -R, --recursive    ", L"Recursively touch files in specified directory and" );
	printOptionUsage( L"                       ", L"all subdirectories." );
	printOptionUsage( L"-v, --verbose          ", L"Output the result of every file processed." );
	printOptionUsage( L"-x, --creation-time    ", L"Change file creation time." );
	std::wcout << std::endl;
	printOptionUsage( L"-r, --reference REFFILE", L"Use this file's times instead of current time." );
	printOptionUsage( L"-t, --time TIME        ", L"Use [[CC]YY]MMDDhhmm[.ss] instead of current time." );
	printOptionUsage( L"-d, --date DATETIME    ", L"Use YYYY-MM-DDThh:mm:SS[.frac] instead of");
	printOptionUsage( L"                       ", L"current time." );
	std::wcout << std::endl;
	printOptionUsage( L"-h, --help             ", L"Display this help and exit." );
	printOptionUsage( L"--version              ", L"Display version information and exit." );

	std::wcout << std::endl;
	std::wcout << L"Note that the -d and -t options accept different time-date formats." << std::endl;
	std::wcout << std::endl;
	std::wcout << L"Mandatory arguments to long options are mandatory for short options too." << std::endl;
	std::wcout << std::endl;
	std::wcout << L"http://www.touchdotexe.com" << std::endl;
}
//------------------------------------------------------------------------------------------------//
void printOptionUsage( const wchar_t* option, const wchar_t* desc )
{
	const wchar_t kDoubleSpace[] = L"  ";
	std::wcout << kDoubleSpace << option << kDoubleSpace << desc << std::endl;
}
//------------------------------------------------------------------------------------------------//
void printVersion( void )
{
	std::wcout << L"touch.exe version " << TOUCH_EXE_MAJOR_VERSION << L"." << TOUCH_EXE_MINOR_VERSION << std::endl;
	std::wcout << L"Written by Stephane Duguay <s@binarez.com>" << std::endl;
	std::wcout << L"This software is provided as is, for free and without warranty." << std::endl;
	std::wcout << L"Go to http://www.touchdotexe.com to access source code." << std::endl;
}
//------------------------------------------------------------------------------------------------//
void printStrings( const std::vector< std::wstring >& toPrint )
{
	std::vector< std::wstring >::const_iterator iter = toPrint.begin();
	const std::vector< std::wstring >::const_iterator iterEnd = toPrint.end();
	for( ; iter != iterEnd; ++iter )
	{
		printUnicodeSafely( *iter );
		std::wcout << std::endl;
	}
}
//------------------------------------------------------------------------------------------------//
void printClosingMessage( bool gotSuccess, const std::vector< std::wstring >& errors, int exitCode )
{
	if( !gotSuccess )
	{
		std::wcout << L"None, sorry!" << std::endl;
	}
	
	if( !errors.empty() )
	{
		std::wcerr << std::endl << L"Errors:" << std::endl;
		printStrings( errors );
	}
	std::wcout << std::endl << L"Exit code: " << exitCode << std::endl;
}
//------------------------------------------------------------------------------------------------//
void printUnicodeSafely( const std::wstring& filename )
{
	std::wstring::const_iterator iter = filename.begin();
	const std::wstring::const_iterator iterEnd = filename.end();
	for( ; iter != iterEnd; ++iter )
	{
		try
		{
			std::wcout << *iter;
		}
		catch( const std::ios_base::failure& )
		{
			std::wcout.clear();
			std::wcout << kQuestionMark;
		}
	}
}
//------------------------------------------------------------------------------------------------//
void printTimeLoadErrorMsg( TouchTime::eLoadResult timeLoadResult )
{
	std::wcout << kProgramName << L": ";
	switch( timeLoadResult )
	{
	case TouchTime::ErrMultiSources:
		std::wcout << L"multiple time sources specified. Pick one!" << std::endl;
		break;
	case TouchTime::ErrInvalidRefFile:
		std::wcout << L"cannot access reference file." << std::endl;
		break;
	case TouchTime::ErrInvalidTimestamp:
		std::wcout << L"invalid -t option argument." << std::endl;
		std::wcout << L"Expected format: [[CC]YY]MMDDhhmm[.ss]" << std::endl;
		break;
	case TouchTime::ErrInvalidDateTime:
		std::wcout << L"invalid -d option argument." << std::endl;
		std::wcout << "Expected format: YYYY-MM-DDThh:mm:SS[.frac]" << std::endl;
		break;
	case TouchTime::ErrSystemTimeFailure:
		std::wcout << L"system time error." << std::endl;
		break;
	default:
		std::wcout << L"unknown error." << std::endl;
		break;
	}

	std::wcout << L"Type '" << kProgramName << " --help' for more information." << std::endl;
}
