/*
Copyright (c) 2010, Stephane Duguay < s (@) binarez (.) com >
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this list of conditions and
the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions
and the following disclaimer in the documentation and/or other materials provided with the distribution.
* Neither the name of the <ORGANIZATION> nor the names of its contributors may be used to endorse or promote
products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "touchtime.h"
#include "touchoptions.h"
#include "touchutils.h"
//------------------------------------------------------------------------------------------------//
TouchTime::TouchTime( void )
{
	reset();
}
//------------------------------------------------------------------------------------------------//
TouchTime::~TouchTime( void )
{
}
//------------------------------------------------------------------------------------------------//
void TouchTime::reset( void )
{
	mCreationTime.dwLowDateTime		= 0;
	mCreationTime.dwHighDateTime	= 0;
	mAccessTime.dwLowDateTime		= 0;
	mAccessTime.dwHighDateTime		= 0;
	mModifTime.dwLowDateTime		= 0;
	mModifTime.dwHighDateTime		= 0;
}
//------------------------------------------------------------------------------------------------//
TouchTime::eLoadResult TouchTime::loadFromOptions( const TouchOptions& touchOptions )
{
	reset();

	const bool hasRefFile = !touchOptions.mReferenceFile.empty();
	const bool hasTimeStamp = !touchOptions.mTimestamp.empty();
	const bool hasDateTime = !touchOptions.mDateTime.empty();

	// It has none or one of the three.
	if( (hasRefFile || hasTimeStamp || hasDateTime)		&&
		!TouchUtils::tripleLogicalXOR( hasRefFile, hasTimeStamp, hasDateTime ) )
	{
		return TouchTime::ErrMultiSources;
	}

	if( hasRefFile )
	{
		if( !loadFromReferenceFile( touchOptions.mReferenceFile ) )
			{ return TouchTime::ErrInvalidRefFile; }
	}
	else if( hasTimeStamp )
	{
		if( !loadFromTimestamp( touchOptions.mTimestamp ) )
			{ return TouchTime::ErrInvalidTimestamp; }
	}
	else if( hasDateTime )
	{
		if( !loadFromDateTime( touchOptions.mDateTime ) )
			{ return TouchTime::ErrInvalidDateTime; }
	}
	else	// Current time
	{
		SYSTEMTIME sysTime;
		GetSystemTime( &sysTime );
		if( SystemTimeToFileTime( &sysTime, &mAccessTime ) )
		{
			mModifTime = mAccessTime;
			mCreationTime = mAccessTime;
		}
		else
		{
			return TouchTime::ErrSystemTimeFailure;
		}
	}

	return TouchTime::Success;
}
//------------------------------------------------------------------------------------------------//
bool TouchTime::loadFromReferenceFile( const std::wstring& absRefFile )
{
	const HANDLE hRefFile = CreateFile( absRefFile.c_str(), GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
	if( hRefFile == INVALID_HANDLE_VALUE )
	{ return false; }

	const bool gotTime = GetFileTime( hRefFile, &mCreationTime, &mAccessTime, &mModifTime ) != 0;

	CloseHandle( hRefFile );
	return gotTime;
}
//------------------------------------------------------------------------------------------------//
bool TouchTime::loadFromTimestamp( const std::wstring& origTimestamp )
{
	// Format = [[CC]YY]MMDDhhmm[.ss]

	if( origTimestamp.length() < 8 )	// Minimal length
		{ return false; }

	std::wstring timestamp = origTimestamp;
	SYSTEMTIME sysTimestamp = { 0 };

	// Seconds
	const std::wstring::size_type dotPos = timestamp.rfind( '.' );
	if( dotPos != std::wstring::npos && dotPos < (timestamp.length() - 1) )
	{
		const std::wstring seconds = timestamp.substr( dotPos + 1 );
		sysTimestamp.wSecond = seconds.empty() ? 0 : static_cast< WORD >( _wtoi( seconds.c_str() ) );
	}

	// Cut off [.ss]
	timestamp = timestamp.substr( 0, dotPos );

	// Year
	switch( timestamp.length() )
	{
	case 12:
		sysTimestamp.wYear = static_cast< WORD >( _wtoi( timestamp.substr( 0, 4 ).c_str() ) );
		timestamp = timestamp.substr( 4 );
		break;
	case 10:
		sysTimestamp.wYear = static_cast< WORD >( _wtoi( timestamp.substr( 0, 2 ).c_str() ) );
		timestamp = timestamp.substr( 2 );
		// More info at http://www.opengroup.org/onlinepubs/9699919799/utilities/touch.html
		if( sysTimestamp.wYear >= 0 && sysTimestamp.wYear <= 68 )
		{ sysTimestamp.wYear += 2000; }
		else if( sysTimestamp.wYear >= 69 && sysTimestamp.wYear <= 99 )
		{ sysTimestamp.wYear += 1900; }
		else // Very bad!
		{ return false; }
		break;
	case 8:
		{
			SYSTEMTIME currentTime;
			GetSystemTime( &currentTime );
			sysTimestamp.wYear = currentTime.wYear;
		}
		break;
	default:
		return false;
	}

	if( timestamp.length() != 8 )
		{ return false; }
	sysTimestamp.wMonth = static_cast< WORD >( _wtoi( timestamp.substr( 0, 2 ).c_str() ) );
	sysTimestamp.wDay = static_cast< WORD >( _wtoi( timestamp.substr( 2, 2 ).c_str() ) );
	sysTimestamp.wHour = static_cast< WORD >( _wtoi( timestamp.substr( 4, 2 ).c_str() ) );
	sysTimestamp.wMinute = static_cast< WORD >( _wtoi( timestamp.substr( 6, 2 ).c_str() ) );
	sysTimestamp.wMilliseconds = 0;

	// This conversion also validates that the time is valid
	FILETIME temp;
	if( SystemTimeToFileTime( &sysTimestamp, &temp ) )
	{
		LocalFileTimeToFileTime( &temp, &mAccessTime );
		mModifTime = mAccessTime;
		mCreationTime = mAccessTime;
		return true;
	}

	return false;
}
//------------------------------------------------------------------------------------------------//
bool TouchTime::loadFromDateTime( const std::wstring& dateTime )
{
	// Format = YYYY-MM-DDThh:mm:SS[.frac]
	if( dateTime.length() < 19 )
		{ return false; }

	// Validate format
	if(	dateTime[4] != '-'	||
		dateTime[7] != '-'	||
		( dateTime[10] != 'T' && dateTime[10] != ' ' ) ||
		dateTime[13] != ':' ||
		dateTime[16] != ':' )
	{
		return false;
	}

	SYSTEMTIME sysTimestamp = { 0 };
	sysTimestamp.wYear	 = static_cast< WORD >( _wtoi( dateTime.substr( 0,  4 ).c_str() ) );
	sysTimestamp.wMonth  = static_cast< WORD >( _wtoi( dateTime.substr( 5,  2 ).c_str() ) );
	sysTimestamp.wDay	 = static_cast< WORD >( _wtoi( dateTime.substr( 8,  2 ).c_str() ) );
	sysTimestamp.wHour	 = static_cast< WORD >( _wtoi( dateTime.substr( 11, 2 ).c_str() ) );
	sysTimestamp.wMinute = static_cast< WORD >( _wtoi( dateTime.substr( 14, 2 ).c_str() ) );
	sysTimestamp.wSecond = static_cast< WORD >( _wtoi( dateTime.substr( 17, 2 ).c_str() ) );

	sysTimestamp.wMilliseconds = 0;
	if( dateTime.length() > 19 )	// OK, we have a fraction to parse
	{
		if( dateTime[19] != '.' && dateTime[19] != ',' )	// Bad fractional format
		{
			return false;
		}
		else if( dateTime.length() > 20 )	// Ok there's a fraction to read
		{
			static const std::wstring kZeroDot = L"0.";
			double frac = _wtof( (kZeroDot + dateTime.substr( 20 )).c_str() );
			sysTimestamp.wMilliseconds = static_cast< WORD >( frac * 1000 + 0.5 );
		}
	}

	// This conversion also validates that the time is valid
	FILETIME temp;
	if( SystemTimeToFileTime( &sysTimestamp, &temp ) )
	{
		LocalFileTimeToFileTime( &temp, &mAccessTime );
		mModifTime = mAccessTime;
		mCreationTime = mAccessTime;
		return true;
	}

	return false;
}
//------------------------------------------------------------------------------------------------//
